/***************************************************************************
 *   Copyright (C) 2020 by santiago González                               *
 *   santigoro@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, see <http://www.gnu.org/licenses/>.  *
 *                                                                         *
 ***************************************************************************/

#ifndef EMCU_H
#define EMCU_H

#include "e-element.h"
#include "itemlibrary.h"
#include "mcuport.h"
#include "mcutimer.h"
#include "mcuuart.h"
#include "mcuinterrupts.h"
#include "mcudataspace.h"
#include "mcuinterface.h"

class McuCore;
class McuTimer;
class McuWdt;
class McuSleep;

enum{
    R_READ = 0,
    R_WRITE,
};

enum mcuState_t{
    mcuStopped=0,
    mcuRunning,
    mcuSleeping
};

class McuVref;

class eMcu : public McuInterface, public DataSpace
{
        friend class McuCreator;
        friend class McuCore;
        friend class Mcu;

    public:
        eMcu( QString id );
        ~eMcu();

        virtual void initialize() override;
        virtual void runEvent() override;

        // Overrides of McuInterface-----------------------------------
        //-------------------------------------------------------------
        virtual void stepCpu() override;

        virtual uint8_t  getRamValue( int address ) override;
        virtual void     setRamValue( int address, uint8_t value ) override;
        virtual uint16_t getFlashValue( int address ) override { return m_progMem[address]; }
        virtual void     setFlashValue( int address, uint16_t value ) override { m_progMem[address] = value; }
        virtual uint8_t  getRomValue( int address ) override { return m_eeprom[address]; }
        virtual void     setRomValue( int address, uint8_t value ) override { m_eeprom[address] = value; }

        virtual uint16_t getRegAddress( QString reg ) override;  // Get Reg address by name

        virtual int status() override;
        virtual int pc() override;
        virtual uint64_t cycle() override { return m_cycle; }
        //-------------------------------------------------------------
        // End Overrides of McuInterface-------------------------------

        void cpuReset( bool reset );
        void sleep( bool s );

        double freqMHz() { return m_freq*1e-6; }
        void setFreq( double freq );
        uint64_t psPerInst() { return m_psPerInst; }  // picoseconds per instruction cycle
        uint64_t psPerClk() { return m_psPerClk; }
        //double cpi() { return m_cPerInst; }       // Clock ticks per Instruction Cycle

        McuTimer* getTimer( QString name ) { return m_timers.getTimer( name ); }
        McuPort* getPort( QString name );
        McuPin*  getPin( QString pinName );

        McuWdt* watchDog() { return m_wdt; }
        McuVref* vrefModule();
        //McuSleep* sleepModule();

        void wdr();

        void enableInterrupts( uint8_t en );

        bool setCfgWord( uint16_t addr, uint16_t data );
        uint16_t getCfgWord( uint16_t addr=0 );

        McuCore* cpu;
        int cyclesDone;

    protected:
        mcuState_t m_state;

        uint64_t m_cycle;
        std::vector<uint16_t> m_progMem;  // Program memory
        QHash<uint16_t, uint16_t> m_cfgWords; // Config words

        Interrupts m_interrupts;
        McuTimers  m_timers;
        QHash<QString, McuPort*>  m_mcuPorts; // Access PORTS by name
        std::vector<McuModule*> m_modules;
        std::vector<McuUsart*> m_usarts;

        McuSleep* m_sleepModule;
        McuVref* m_vrefModule;
        McuWdt* m_wdt;

        bool m_resetState;

        double m_freq;           // Clock Frequency in MegaHerzs
        double m_cPerInst;       // Clock ticks per Instruction Cycle
        uint64_t m_psPerInst;    // picoseconds per Instruction Cycle
        uint64_t m_psPerClk;     // picoseconds per Clock Cycle
};

#endif
